#include <stdarg.h>
#include <time.h>

#include "pstruct.h"

void command_reset_session(character *, char *, const char *, void *);
void command_comment(character *, char *, const char *, void *);   
void command_comments(character *, char *, const char *, void *);   
void command_session(character *, char *, const char *, void *);   

void onlogin(character *c, void *s);

#define LEN 80
#define EXPIRE_TIME 5*60 /* sessions can be over written in 5 minutes */

/* the hideous globals */
char session[LEN] = "";            /* the text of the session */
char setter[MAX_NAME_LENGTH] = ""; /* name of the person who set the session */
time_t set_time = 0; /* the time at which the session comment is set */

struct comment {
  char comment[LEN];
};

command clist[] = {
  { "comment",       "",   command_comment,       "n_base" },
  { "comments",      "",   command_comments,      "n_base" },
  { "session",       "",   command_session,       "n_base" },
  { "reset_session", "",   command_reset_session, "s_base" },
  { NULL,            NULL, NULL,                  NULL     }
};

unsigned pversion = 1;

struct pheader p = {
  "Session Comments",
  1,
  "Flibble",
  clist,
  sizeof(struct comment),
  NULL,
  NULL,
  NULL,
  NULL,
  onlogin,
  NULL,
  NULL,
  NULL,
  "comment.hlp"
}; 

void onlogin(character *c, void *s)
{
  struct comment *co = s;
  
  STRNCOPY(co->comment, "", sizeof(co->comment));
}

void command_reset_session(character *c, char *r, const char *comm, void *s)
{
  listnode *pNode = AllPlayers.head.next;
  descriptor *des = getdes(c);
  size_t offset = ((char *)s) - ((char *)des);

  /* Blank the session */  
  STRNCOPY(session, "", sizeof(session));
  STRNCOPY(setter, "", sizeof(setter));
  set_time = time(NULL) - EXPIRE_TIME - 1;

  /* blank everyones previous comments */
  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode,character *, characterlink);
    descriptor *d = getdes(pChar);
    void *so = ((char *)d) + offset;
    struct comment *co = so;

    STRNCOPY(co->comment, "", sizeof(co->comment));

    pNode = pNode->next;
  }

  send_to_char(c, "You reset the session\n");
  send_to_all("^B[Session] The Session is reset\n");
}

void command_comments(character *c, char *r, const char *comm, void *s)
{
  listnode *pNode = AllPlayers.head.next;
  descriptor *des = getdes(c);
  size_t offset = ((char *)s) - ((char *)des);
  char temp[LEN + 9];
  int num = 0;

  if(STREQ(session, "")) {
    send_to_char(c, "No one has set a session yet\n");
    return;
  }

  STRNCOPY(temp, "Session: ", sizeof(temp));
  STRNAPPEND(temp, session, sizeof(temp));

  bar2(des, temp, NULL, NULL);     
  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode,character *, characterlink);
    descriptor *d = getdes(pChar);
    void *so = ((char *)d) + offset;
    struct comment *co = so;
    if(!STREQ(co->comment, "")) {
      send_to_char(c, "%-*s - %s\n", MAX_NAME_LENGTH, GETNAME(pChar),
                                      co->comment);
      num++;
    }
    pNode = pNode->next;
  }
  if(num == 0) {
    send_to_char(c, "No one has commented yet\n");
  }

  bar2(des, NULL, NULL, NULL);     


}

void command_comment(character *c, char *r, const char *comm, void *s)
{
  struct comment *co = s;

  if(STREQ(session, "")) {
    send_to_char(c, "No one has set a session yet\n");
    return;
  }

  if(STREQ(r, "")) {
    send_to_char(c, "What do you wish to set your session comment too?\n");
    return;
  }

  STRNCOPY(co->comment, r, sizeof(co->comment));
  send_to_char(c, "You set your session comment to '%s'\n", co->comment);
  send_to_all("^B[Session] %s comments^n\n", GETNAME(c));

}

void command_session(character *c, char *r, const char *comm, void *s)
{
  listnode *pNode = AllPlayers.head.next;
  descriptor *des = getdes(c);
  size_t offset = ((char *)s) - ((char *)des);

  /* not argument, they want details of who set and when the session was set */
  if(STREQ(r, "")) {
    if(STREQ(session, "")) {
      send_to_char(c, "No one has set a session yet, why don't you?\n");
    } else {
      send_to_char(c, "The current session '%s' was set by %s\n", session,
                   setter);
      if(time(NULL) < set_time + EXPIRE_TIME) {
        send_to_char(c, "A new session can be set in %d seconds\n",
                     set_time + EXPIRE_TIME - time(NULL));
      } else {
        send_to_char(c, "A new session can be set *now*\n");
      }
    }
    return;
  }

  /* Has the old session expired yet ? */
  if(time(NULL) < set_time + EXPIRE_TIME) {
    send_to_char(c, "You cannot set a new session yet, one can be set in %d seconds\n",
                 set_time + EXPIRE_TIME - time(NULL));
    return;
  }

  /* Actual set up the new session */  
  STRNCOPY(session, r, sizeof(session));
  STRNCOPY(setter, GETNAME(c), sizeof(setter));
  set_time = time(NULL);

  /* blank everyones previous comments */
  while (LIST_NODE_IS_REAL(pNode)) {
    character *pChar = LIST_GET_DATA(pNode,character *, characterlink);
    descriptor *d = getdes(pChar);
    void *so = ((char *)d) + offset;
    struct comment *co = so;

    STRNCOPY(co->comment, "", sizeof(co->comment));

    pNode = pNode->next;
  }

  send_to_char(c, "You the session to be '%s'\n", session);
  send_to_all("^B[Session] %s sets a new session^n\n", GETNAME(c));
}
